/*
 *  AMT IDE-R demo.
 *
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <pthread.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <sys/eventfd.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <poll.h>
#include <fcntl.h>
#include <setjmp.h>
#include <stdio.h>
#include "ider.h"
#include "rdwr.h"

static pthread_mutex_t mtx=PTHREAD_MUTEX_INITIALIZER;

static int efd;
static int diskfd=-1;
static int cddvdfd=-1;
static jmp_buf jev;

static void handle_regs(struct iderctx *ctx)
{
	int state;
	char bfr[128];

	printf("\nide register menu:\n");
	state=ider_regs(ctx,REGS_STATE);
	printf("current register state: ");
	switch(state)
	{
	case REGS_UNKNOWN:
		printf("unknown\n");
		break;
	case REGS_NONE:
		printf("not available\n");
		break;
	case REGS_DISABLED:
		printf("disabled\n");
		printf("1 enable register now\n");
		printf("2 enable register on reboot\n");
		break;
	case REGS_ENABLED:
		printf("enabled\n");
		printf("1 disable register now\n");
		printf("2 disable register graceful\n");
		printf("3 disable register on reboot\n");
		break;
	}

	printf("0 main menu\n");
	printf("enter choice: ");
	fflush(stdout);
	if(!fgets(bfr,sizeof(bfr),stdin))return;
	if(bfr[1]=='\n')bfr[1]=0;
	if(!bfr[1])
	{
		pthread_mutex_lock(&mtx);
		if(state==REGS_ENABLED)switch(bfr[0])
		{
		case '1':
			ider_regs(ctx,REGS_DISABLE_NOW);
			break;
		case '2':
			ider_regs(ctx,REGS_DISABLE_GRACEFUL);
			break;
		case '3':
			ider_regs(ctx,REGS_DISABLE_REBOOT);
			break;
		}
		else if(state==REGS_DISABLED)switch(bfr[0])
		{
		case '1':
			ider_regs(ctx,REGS_ENABLE_NOW);
			break;
		case '2':
			ider_regs(ctx,REGS_ENABLE_REBOOT);
			break;
		}
		pthread_mutex_unlock(&mtx);
	}
}

static void handle_disk(struct iderctx *ctx)
{
	int imode;
	int omode;
	char bfr[128];

	printf("\ndisk menu:\n");
	printf("disk image is currently %s\n",
		diskfd!=-1?"attached":"detached");
	if(diskfd==-1)
	{
		printf("1 attach disk image rw\n");
		printf("2 attach disk image ro\n");
	}
	else printf("1 detach disk image\n");

	printf("0 main menu\n");
	printf("enter choice: ");
	fflush(stdout);
	if(!fgets(bfr,sizeof(bfr),stdin))return;
	if(bfr[1]=='\n')bfr[1]=0;
	if(!bfr[1])
	{
		if(diskfd==-1)switch(bfr[0])
		{
		case '1':
			imode=DEV_RW;
			omode=O_RDWR;
			goto common;

		case '2':
			imode=DEV_RO;
			omode=O_RDONLY;
common:			printf("enter path: ");
			if(!fgets(bfr,sizeof(bfr),stdin))return;
			if(*bfr)if(bfr[strlen(bfr)-1]=='\n')
				bfr[strlen(bfr)-1]=0;
			if(!*bfr)return;
			if((diskfd=open(bfr,omode))==-1)
			{
				perror("open");
				return;
			}
			if(ider_device(ctx,DEV_FLOPPY,imode,diskfd))
			{
				close(diskfd);
				diskfd=-1;
				printf("failed\n");
			}
			break;
		}
		else if(bfr[0]=='1')
		{
			ider_device(ctx,DEV_FLOPPY,DEV_NONE,-1);
			close(diskfd);
			diskfd=-1;
		}
	}
}

static void handle_cddvd(struct iderctx *ctx)
{
	char bfr[128];

	printf("\ncd/dvd menu:\n");
	printf("cd/dvd image is currently %s\n",
		cddvdfd!=-1?"attached":"detached");
	if(cddvdfd==-1)printf("1 attach cd/dvd image ro\n");
	else printf("1 detach cd/dvd image\n");

	printf("0 main menu\n");
	printf("enter choice: ");
	fflush(stdout);
	if(!fgets(bfr,sizeof(bfr),stdin))return;
	if(bfr[1]=='\n')bfr[1]=0;
	if(!bfr[1])
	{
		if(cddvdfd==-1)switch(bfr[0])
		{
		case '1':
			printf("enter path: ");
			if(!fgets(bfr,sizeof(bfr),stdin))return;
			if(*bfr)if(bfr[strlen(bfr)-1]=='\n')
				bfr[strlen(bfr)-1]=0;
			if(!*bfr)return;
			if((cddvdfd=open(bfr,O_RDONLY))==-1)
			{
				perror("open");
				return;
			}
			if(ider_device(ctx,DEV_CDDVD,DEV_RO,cddvdfd))
			{
				close(cddvdfd);
				cddvdfd=-1;
				printf("failed\n");
			}
			break;
		}
		else if(bfr[0]=='1')
		{
			ider_device(ctx,DEV_CDDVD,DEV_NONE,-1);
			close(cddvdfd);
			cddvdfd=-1;
		}
	}
}

static void *processor(void *data)
{
	uint64_t dummy=1;
	struct iderctx *ctx=data;
	sigset_t set;

	sigfillset(&set);
	pthread_sigmask(SIG_BLOCK,&set,NULL);

	pthread_mutex_lock(&mtx);

	while(1)if(ider_engine(ctx))break;

	pthread_mutex_unlock(&mtx);

	dummy=write(efd,&dummy,sizeof(dummy));

	pthread_exit(NULL);
}

static void handler(int unused)
{
	uint64_t dummy=1;
	dummy=write(efd,&dummy,sizeof(dummy));
	printf("\n");
	longjmp(jev,1);
}

int main(int argc,char *argv[])
{
	int r=1;
	int fd;
	int x;
	int debug=0;
	int verbose=0;
	int v4flag=0;
	int persist=0;
	char *addr;
	char *user;
	char *pass;
	struct iderctx *ctx;
	struct ctx c;
	uint64_t dummy=1;
	pthread_t h;
	char bfr[128];

	c.iodebug=0;
	c.cacert=NULL;
	c.lck=&mtx;
	c.lock=(void (*)(void *))pthread_mutex_lock;
	c.unlock=(void (*)(void *))pthread_mutex_unlock;

	if(!(user=getenv("IDER_USER")))user="admin";
	pass=getenv("IDER_PASS");

	while((x=getopt(argc,argv,"dv4u:p:P"))!=-1)switch(x)
	{
	case 'd':
		debug++;
		break;

	case 'v':
		verbose=1;
		break;

	case '4':
		v4flag=1;
		break;

	case 'u':
		user=optarg;
		break;

	case 'p':
		pass=optarg;
		break;

	case 'P':
		persist=1;
		break;
	}

	if(optind!=argc-1)
	{
		fprintf(stderr,"missing target host, bailing out\n");
		goto err1;
	}
	else addr=argv[optind];

	if(!pass)
	{
		fprintf(stderr,"missing password, bailing out\n");
		goto err1;
	}

	if(debug>1)c.iodebug=1;

	if((efd=c.efd=eventfd(0,0))==-1)
	{
		perror("eventfd");
		goto err1;
	}

	signal(SIGINT,handler);
	signal(SIGHUP,handler);
	signal(SIGTERM,handler);
	signal(SIGQUIT,handler);
	signal(SIGPIPE,SIG_IGN);

	if(doconnect(&c,addr,v4flag))goto err2;

	if(!(ctx=ider_init(persist,verbose,debug,60000,0,0,0,
		io,iowait,NULL,&fd)))goto err3;

	if(ider_start(ctx,user,pass))goto err4;

	if(pthread_create(&h,NULL,processor,ctx))
	{
		perror("pthread_create");
		goto err4;
	}

	if(setjmp(jev))goto out;

	while(1)
	{
		printf("\nmain menu:\n");
		printf("1 ide registers\n");
		printf("2 disk image\n");
		printf("3 cd/dvd image\n");
		printf("0 exit\n");
		printf("enter choice: ");
		fflush(stdout);
		if(!fgets(bfr,sizeof(bfr),stdin))break;
		if(bfr[1]=='\n')bfr[1]=0;
		if(!bfr[1])switch(bfr[0])
		{
		case '0':
			goto out;

		case '1':
			handle_regs(ctx);
			break;

		case '2':
			handle_disk(ctx);
			break;

		case '3':
			handle_cddvd(ctx);
			break;
		}
	}

out:	dummy=write(efd,&dummy,sizeof(dummy));

	pthread_join(h,NULL);

	dummy=read(efd,&dummy,sizeof(dummy));

	if(cddvdfd!=-1)close(cddvdfd);
	if(diskfd!=-1)close(diskfd);

	r=0;

err4:	ider_end(ctx);
err3:	doclose(&c);
err2:	close(efd);
err1:	return r;
}
